from ..browser.get_current_url import get_current_url
from ..browser.get_current_page_title import get_current_page_title

class WaitForElementTimeoutException(Exception):
    def __init__(self, driver: object, xpath: str) -> None:
        current_url = get_current_url(driver)
        self.message = f"On page {current_url}, waiting for element timed out: {xpath}"
        super().__init__(self.message)

    def __str__(self) -> str:
        return self.message

class WaitForPageTitleToChangeTimeoutException(Exception):
    def __init__(self, driver: object, page_title_or_fragment: str) -> None:
        current_url = get_current_url(driver)
        current_page_title = get_current_page_title(driver)
        self.message = f"On page {current_url} with title \"{current_page_title}\", check of page title fragment timed out, or the title fragment doesn't match: {page_title_or_fragment}"
        super().__init__(self.message)

    def __str__(self) -> str:
        return self.message

class WaitForUrlTimeoutException(Exception):
    def __init__(self, driver: object, url: str) -> None:
        current_url = get_current_url(driver)
        self.message = f"On page {current_url}, waiting for URL timed out, or the URL fragment doesn't exist: {url}"
        super().__init__(self.message)

    def __str__(self) -> str:
        return self.message



from enum import Enum, unique

@unique
class BrowserType(Enum):
    """Class to define browser type, e.g. Chrome, Firefox, etc."""
    
    CHROME = "Chrome"
    EDGE = "Edge"
    FIREFOX = "Firefox"
    INTERNET_EXPLORER = "Internet Explorer"
    OPERA = "Opera"
    SAFARI = "Safari"



from selenium.common.exceptions import TimeoutException, NoSuchElementException
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait
from .check_if_does_element_exist import check_if_does_element_exist
from .wait_for_element import wait_for_element
from ..constant import timeout
from ..model.browser.base.driver import BrowserDriver
from ..model.driver_methods import DriverMethods

def check_if_is_element_clickable(driver: object, xpath: str, timeout: int = timeout.DEFAULT) -> bool:
    wait_for_element(driver, xpath, timeout)
    try:
        element = WebDriverWait(driver, 0.5).until(EC.element_to_be_clickable((By.XPATH, xpath)))
        return element is not None
    except TimeoutException:
        return False
    except NoSuchElementException:
        return False

def check_if_is_element_enabled(driver: object, xpath: str) -> bool:
    wait_for_element(driver, xpath)
    try:
        element = driver.find_element_by_xpath(xpath)
        return element.is_enabled()
    except NoSuchElementException:
        return False

def check_if_is_element_disabled(driver: object, xpath: str) -> bool:
    return not check_if_is_element_enabled(driver, xpath)

def check_if_is_element_visible(driver: object, xpath: str) -> bool:
    try:
        element = driver.find_element_by_xpath(xpath)
        return element.is_displayed()
    except NoSuchElementException:
        return False

def check_if_is_image_element_loaded(driver: object, element: object) -> bool:
    is_image_loaded = driver.execute_script(
        "return arguments[0].complete && typeof arguments[0].naturalWidth != 'undefined' && arguments[0].naturalWidth > 0;",
        element)
    return is_image_loaded

def check_if_is_image_loaded(driver: object, xpath: str, timeout: int = timeout.DEFAULT) -> bool:
    wait_for_element(driver, xpath, timeout)
    element = driver.find_element_by_xpath(xpath)
    return check_if_is_image_element_loaded(driver, element)

class CheckIfDriverMethods(DriverMethods):
    def __init__(self, browser_driver: BrowserDriver) -> None:
        super().__init__(browser_driver)

    def does_element_exist(self, xpath: str,) -> bool:
        """Check if element exists."""

        return check_if_does_element_exist(self._driver, xpath)

    def is_element_clickable(self, xpath: str, timeout: int = timeout.DEFAULT) -> bool:
        """Check if element is clickable."""

        return check_if_is_element_clickable(self._driver, xpath, timeout)

    def is_element_enabled(self, xpath: str) -> bool:
        """Check whether element is enabled."""

        return check_if_is_element_enabled(self._driver, xpath)

    def is_element_disabled(self, xpath: str) -> bool:
        """Check whether element is disabled."""

        return check_if_is_element_disabled(self._driver, xpath)

    def is_element_visible(self, xpath: str) -> bool:
        """Check visibility status of an element."""

        return check_if_is_element_visible(self._driver, xpath)

    def is_image_loaded(self, xpath: str, timeout: int = timeout.DEFAULT) -> bool:
        """Check is image is loaded and ready in the DOM."""

        return check_if_is_image_loaded(self._driver, xpath, timeout)

    def is_image_element_loaded(self, element: object) -> bool:
        """Check is image element is loaded and ready in the DOM."""

        return check_if_is_image_element_loaded(self._driver, element)



from dataclasses import dataclass
from typing import Union
from ..constant import timeout

@dataclass
class CookieBannerSettings:
    """Object with data needed to accept or decline cookies from a banner.
    
    button_xpath: Can be for either accept of decline cookies.
    
    url: Optional URL from where to handle the cookie banner.
    
    has_loaded_wait_seconds: Minor grace time to make sure the cookie banner has loaded.
    
    has_loaded_xpath: Check if banner element has loaded so the cookie banner is ready.
    
    has_disappeared_wait_seconds: Minor grace time to make sure the cookie banner has disappeared and that the cookie information has been saved before proceeding."""

    button_xpath: str
    url: Union[str, None] = None
    has_loaded_wait_seconds: int = timeout.DEFAULT
    has_loaded_xpath: Union[str, None] = None
    has_disappeared_wait_seconds: int = timeout.SHORT



from .browser.base.driver import BrowserDriver

class DriverMethods:
    """Super class with initializer that can be extended in sub classes for web driver methods."""

    def __init__(self, browser_driver: BrowserDriver) -> None:
        self._browser_driver: BrowserDriver = browser_driver
        self._driver: object = browser_driver.webdriver



from enum import Enum, unique

@unique
class PageLoadStrategy(Enum):
    """Class to configure page load strategy for Selenium.

    NORMAL: Used by default. Waits for all resources to download. Ready state: Complete.

    EAGER: DOM access is ready, but other resources like images may still be loading. Ready state: Interactive.

    NONE: Does not block web driver at all. Ready state: Any."""
    
    NORMAL = "normal"    
    EAGER = "eager"
    NONE = "none"



from typing import List
from selenium.common.exceptions import TimeoutException, NoSuchElementException
from .get_current_page_title import get_current_page_title
from .get_current_url import get_current_url
from .wait_for_element import wait_for_element
from .. import helper
from ..constant import timeout
from ..exception.element import NoElementDimensionsFoundException
from ..exception.timeout import WaitForElementTimeoutException
from ..model.browser.base.driver import BrowserDriver
from ..model.driver_methods import DriverMethods

def get_dimensions_of_element(driver: object, xpath: str, timeout: int = timeout.DEFAULT) -> tuple[int, int]:
    wait_for_element(driver, xpath, timeout)
    try:
        dimensions = driver.find_element_by_xpath(xpath).size # Returns dictionary object, e.g. {'height': 598, 'width': 479}.
        return dimensions.get("width"), dimensions.get("height")
    except TimeoutException:
        raise WaitForElementTimeoutException(driver, xpath)
    except NoSuchElementException:
        raise NoElementDimensionsFoundException(driver, xpath)

def get_text_from_element(driver: object, xpath: str, timeout: int = timeout.DEFAULT) -> str:
    def get_inner_text_of_element(driver: object, xpath: str) -> str:
        return driver.find_element_by_xpath(xpath).text

    wait_for_element(driver, xpath, timeout)
    return helper.driver.retry_and_get_text_from_element(get_inner_text_of_element(driver, xpath))

def get_texts_from_multiple_elements(driver: object, xpath: str, timeout: int = timeout.DEFAULT) -> List[str]:
    wait_for_element(driver, xpath, timeout)
    elements = driver.find_elements_by_xpath(xpath)
    return [element.text for element in elements]

def get_url_from_image(driver: object, xpath: str, timeout: int = timeout.DEFAULT) -> str:
    def get_src_attribute_of_element(driver: object, xpath: str) -> str:
        return driver.find_element_by_xpath(xpath).get_attribute("src")

    wait_for_element(driver, xpath, timeout)
    return helper.driver.retry_and_get_text_from_element(get_src_attribute_of_element(driver, xpath))

def get_urls_from_multiple_images(driver: object, xpath: str, timeout: int = timeout.DEFAULT) -> List[str]:
    wait_for_element(driver, xpath, timeout)
    elements = driver.find_elements_by_xpath(xpath)
    return [element.get_attribute("src") for element in elements]

def get_url_from_link(driver: object, xpath: str, timeout: int = timeout.DEFAULT) -> str:
    def get_href_attribute_of_element(driver: object, xpath: str) -> str:
        return driver.find_element_by_xpath(xpath).get_attribute("href")

    wait_for_element(driver, xpath, timeout)
    return helper.driver.retry_and_get_text_from_element(get_href_attribute_of_element(driver, xpath))

def get_urls_from_multiple_links(driver: object, xpath: str, timeout: int = timeout.DEFAULT) -> List[str]:
    wait_for_element(driver, xpath, timeout)
    elements = driver.find_elements_by_xpath(xpath)
    return [element.get_attribute("href") for element in elements]

class GetDriverMethods(DriverMethods):
    def __init__(self, browser_driver: BrowserDriver) -> None:
        super().__init__(browser_driver)
        
    def current_page_title(self) -> str:
        """Get page title of the current page."""

        return get_current_page_title(self._driver)

    def current_url(self) -> str:
        """Get URL of the current page."""

        return get_current_url(self._driver)

    def dimensions_of_element(self, xpath: str, timeout: int = timeout.DEFAULT) -> tuple[int, int]:
        """Get width and height of element in pixels. Usage:

        width, height = browser.get.dimensions_of_element("/element/xpath")"""

        return get_dimensions_of_element(self._driver, xpath, timeout)

    def text_from_element(self, xpath: str, timeout: int = timeout.DEFAULT) -> str:
        """Get text from element.

        This method assumes that the text field shouldn't be empty and therefore will retry to get the text (for better support of single-page apps with extended loading time)."""

        return get_text_from_element(self._driver, xpath, timeout)

    def texts_from_multiple_elements(self, xpath: str, timeout: int = timeout.DEFAULT) -> List[str]:
        """Get array of texts from elements.

        Assumes that the XPath targets multiple elements."""

        return get_texts_from_multiple_elements(self._driver, xpath, timeout)

    def url_from_image(self, xpath: str, timeout: int = timeout.DEFAULT) -> str:
        """Get URL source from image, e.g. <img> tag.

        This method assumes that the image shouldn't be empty and therefore will retry to get the URL (for better support of single-page apps with extended loading time)."""

        return get_url_from_image(self._driver, xpath, timeout)

    def urls_from_multiple_images(self, xpath: str, timeout: int = timeout.DEFAULT) -> List[str]:
        """Get array of URLs from images, e.g. <img> tags.

        Assumes that the XPath targets multiple images."""

        return get_urls_from_multiple_images(self._driver, xpath, timeout)

    def url_from_link(self, xpath: str, timeout: int = timeout.DEFAULT) -> str:
        """Get URL from link, e.g. <a> tag or button.

        This method assumes that the link shouldn't be empty and therefore will retry to get the URL (for better support of single-page apps with extended loading time)."""

        return get_url_from_link(self._driver, xpath, timeout)

    def urls_from_multiple_links(self, xpath: str, timeout: int = timeout.DEFAULT) -> List[str]:
        """Get array of URLs from links, e.g. <a> tags or buttons.

        Assumes that the XPath targets multiple links."""

        return get_urls_from_multiple_links(self._driver, xpath, timeout)



from selenium import webdriver
from .base.driver import BrowserDriver
from .base.type import BrowserType
from ... import helper

class FirefoxBrowserDriver(BrowserDriver):
    def ensure_browser_type(self) -> None:
        self.settings.type = BrowserType.FIREFOX

    def set_webdriver(self) -> object:
        if self.settings.path_to_executable is None:
            return webdriver.Firefox(
                firefox_profile = self.firefox_profile,
                options = self.firefox_options)
        else:
            return webdriver.Firefox(
                executable_path = self.settings.path_to_executable,
                firefox_profile = self.firefox_profile,
                options = self.firefox_options)

    def disable_images(self) -> None:
        if self.settings.disable_images:
            self.firefox_profile.set_preference("permissions.default.image", 2)
            self.firefox_profile.set_preference("dom.ipc.plugins.enabled.libflashplayer.so", "false")
    
    def enable_headless(self) -> None:
        if self.settings.headless:
            self.firefox_options("--headless")

    def set_page_load_strategy(self) -> None:
        self.firefox_options = helper.set.page_load_strategy(self, self.firefox_options)



#from winreg import CloseKey, OpenKey, SetValueEx, HKEY_CURRENT_USER, KEY_ALL_ACCESS, REG_SZ
#TODO: Fix ModuleNotFoundError of winreg
from ..model.browser.base.driver import BrowserDriver
from ..model.browser.base.type import BrowserType

def set_image_loading(browser_driver: BrowserDriver, load_images: bool = True):
    if browser_driver.settings.type is BrowserType.INTERNET_EXPLORER:
        value = "yes" if load_images else "no"
        key = OpenKey(HKEY_CURRENT_USER, r"Software\Microsoft\Internet Explorer\Main", 0, KEY_ALL_ACCESS)
        SetValueEx(key, "Display Inline Images", 0, REG_SZ, value)
        CloseKey(key)

def disable_images(browser_driver: BrowserDriver) -> None:
    set_image_loading(browser_driver, load_images = False)

def enable_images(browser_driver: BrowserDriver) -> None:
    set_image_loading(browser_driver, load_images = True)



from selenium.common.exceptions import TimeoutException, NoSuchElementException
from .wait_for_element import wait_for_element
from ..exception.element import NoElementFoundException
from ..exception.timeout import WaitForElementTimeoutException
from ..model.browser.base.driver import BrowserDriver
from ..model.driver_methods import DriverMethods

def input_clear(driver: object, input_xpath: str) -> None:
    wait_for_element(driver, input_xpath)
    try:
        input_field = driver.find_element_by_xpath(input_xpath)
        input_field.clear()
    except TimeoutException:
        raise WaitForElementTimeoutException(driver, input_xpath)
    except NoSuchElementException:
        raise NoElementFoundException(driver, input_xpath)

def input_value(driver: object, input_xpath: str, value: str) -> None:
    wait_for_element(driver, input_xpath)
    try:
        input_field = driver.find_element_by_xpath(input_xpath)
        input_field.clear() # Always clear input field before entering value
        input_field.send_keys(value)
    except TimeoutException:
        raise WaitForElementTimeoutException(driver, input_xpath)
    except NoSuchElementException:
        raise NoElementFoundException(driver, input_xpath)

class InputDriverMethods(DriverMethods):
    def __init__(self, browser_driver: BrowserDriver) -> None:
        super().__init__(browser_driver)
    
    def clear(self, input_xpath: str) -> None:
        """Clear input form field."""

        input_clear(self._driver, input_xpath)

    def value(self, input_xpath: str, value: str) -> None:
        """Input value into form field."""

        input_value(self._driver, input_xpath, value)



from ..model.browser.base.driver import BrowserDriver
from ..model.driver_methods import DriverMethods

class ScrollDriverMethods(DriverMethods):
    def __init__(self, browser_driver: BrowserDriver) -> None:
        super().__init__(browser_driver)



from dataclasses import dataclass
from typing import Union
from .page_load_strategy import PageLoadStrategy
from .type import BrowserType

@dataclass    
class BrowserSettings:
    """Class to configure the browser driver."""
    
    type: BrowserType = BrowserType.CHROME
    headless: bool = False
    disable_images: bool = False
    page_load_strategy: PageLoadStrategy = PageLoadStrategy.NORMAL
    path_to_executable: Union[str, None] = None



from dataclasses import dataclass
from enum import auto, Enum
from ..constant import timeout

class TimeoutStrategy(Enum):
    """Define timeout strategy.
    
    BREAK: If something times out, stop operation.
    
    CONTINUE: If something times out, continue operation."""
    
    BREAK = auto()
    CONTINUE = auto()

@dataclass
class TimeoutSettings:
    strategy: TimeoutStrategy = TimeoutStrategy.BREAK
    seconds: int = timeout.DEFAULT
    _is_timed_out: bool = False



import subprocess
from ..model.browser.base.driver import BrowserDriver
from ..model.browser.base.type import BrowserType

def set_image_loading(browser_driver: BrowserDriver, load_images: bool = True):
    if browser_driver.settings.type is BrowserType.SAFARI:
        value = "true" if load_images else "false"
        subprocess.call(f"defaults write com.apple.Safari com.apple.Safari.ContentPageGroupIdentifier.WebKit2LoadsImagesAutomatically {value}".split())
        subprocess.call(f"defaults write com.apple.Safari com.apple.Safari.ContentPageGroupIdentifier.WebKit2LoadsImagesAutomatically {value}".split())

def disable_images(browser_driver: BrowserDriver) -> None:
    set_image_loading(browser_driver, load_images = False)

def enable_images(browser_driver: BrowserDriver) -> None:
    set_image_loading(browser_driver, load_images = True)



from ..model.browser.base.type import BrowserType

class LoginException(Exception):
    def __init__(self, username: str) -> None:
        self.message = f"Something went wrong when trying to log in with username {username}. Please try again..."
        super().__init__(self.message)

    def __str__(self) -> str:
        return self.message



__all__ = []

from typing import Union
from .check_if import CheckIfDriverMethods
from .click import ClickDriverMethods
from .combo import ComboDriverMethods
from .get import GetDriverMethods
from .input import InputDriverMethods
from .open import OpenDriverMethods
from .scroll import ScrollDriverMethods
from .select import SelectDriverMethods
from .tools import ToolsDriverMethods
from .validate import ValidateDriverMethods
from .wait import WaitDriverMethods
from .. import helper
from ..model.browser.base.driver import BrowserDriver
from ..model.browser.base.settings import BrowserSettings
from ..model.browser.base.type import BrowserType
from ..model.browser.extension.internet_explorer import InternetExplorerBrowserExtension
from ..model.browser.extension.safari import SafariBrowserExtension

class Browser:
    """Main class of Browserist that sets the Selenium web driver and contains all helper functions."""

    def __init__(self, settings: Union[BrowserSettings, None] = None) -> None:
        """Initiates the browser driver whether the settings calls for Chrome, Firefox, etc."""

        self._browser_driver: BrowserDriver = helper.get.browser_driver(settings)
        self.driver: object = self._browser_driver.webdriver

        if self._browser_driver.settings.type is BrowserType.INTERNET_EXPLORER:
            self.ie: InternetExplorerBrowserExtension = InternetExplorerBrowserExtension(self._browser_driver)
        if self._browser_driver.settings.type is BrowserType.SAFARI:
            self.safari: SafariBrowserExtension = SafariBrowserExtension(self._browser_driver)

        self.check_if: CheckIfDriverMethods  = CheckIfDriverMethods(self._browser_driver)
        self.click:    ClickDriverMethods    = ClickDriverMethods(self._browser_driver)
        self.combo:    ComboDriverMethods    = ComboDriverMethods(self._browser_driver)
        self.get:      GetDriverMethods      = GetDriverMethods(self._browser_driver)
        self.input:    InputDriverMethods    = InputDriverMethods(self._browser_driver)
        self.open:     OpenDriverMethods     = OpenDriverMethods(self._browser_driver)
        self.scroll:   ScrollDriverMethods   = ScrollDriverMethods(self._browser_driver)
        self.select:   SelectDriverMethods   = SelectDriverMethods(self._browser_driver)
        self.tools:    ToolsDriverMethods    = ToolsDriverMethods(self._browser_driver)
        self.validate: ValidateDriverMethods = ValidateDriverMethods(self._browser_driver)
        self.wait:     WaitDriverMethods     = WaitDriverMethods(self._browser_driver)

    def back(self) -> None:
        """Press the browser's back button."""

        self.driver.back()

    def forward(self) -> None:
        """Press the browser's forward button."""

        self.driver.forward()

    def refresh(self) -> None:
        """Refresh the current page."""

        self.driver.refresh()

    def quit(self) -> None:
        """Close the browser."""

        self.driver.quit()



from ..model.browser.base.type import BrowserType

class HeadlessNotSupportedException(Exception):
    def __init__(self, browser_type: BrowserType) -> None:
        self.message = f"{browser_type.value}: This browser doesn't support headless."
        super().__init__(self.message)

    def __str__(self) -> str:
        return self.message



from selenium.common.exceptions import TimeoutException, NoSuchElementException
from .wait_for_element import wait_for_element
from ..constant import timeout
from ..exception.element import NoElementFoundException
from ..exception.timeout import WaitForElementTimeoutException
from ..model.browser.base.driver import BrowserDriver
from ..model.driver_methods import DriverMethods

def click_button(driver: object, xpath: str, timeout: int = timeout.DEFAULT) -> None:
    wait_for_element(driver, xpath, timeout)
    try:
        button = driver.find_element_by_xpath(xpath)
        button.click()
    except TimeoutException:
        raise WaitForElementTimeoutException(driver, xpath)
    except NoSuchElementException:
        raise NoElementFoundException(driver, xpath)

class ClickDriverMethods(DriverMethods):
    def __init__(self, browser_driver: BrowserDriver) -> None:
        super().__init__(browser_driver)

    def button(self, xpath: str, timeout: int = timeout.DEFAULT) -> None:
        """Click button."""

        click_button(self._driver, xpath, timeout)



from .get_current_url import get_current_url
from ..model.browser.base.driver import BrowserDriver
from ..model.driver_methods import DriverMethods

def open_url(driver: object, url: str) -> None:
    driver.get(url)

def open_url_if_not_current(driver: object, url: str, ignore_trailing_slash: bool = True) -> None:
    # TODO: Evaluate without trailing slash
    # TODO: Ignore HTTP?(S) part of URL
    # TODO: Option to ignore parameters in URL
    current_url = get_current_url(driver)
    if current_url != url:
        open_url(driver, url)

def open_url_in_new_tab(driver: object, url: str) -> None:
    driver.execute_script(f"window.open('{url}', '_blank');")
    # TODO: Should the new tab be in focus? Could be an argument, e.g. ... focus_new_tab = False).

class OpenDriverMethods(DriverMethods):
    def __init__(self, browser_driver: BrowserDriver) -> None:
        super().__init__(browser_driver)

    def url(self, url: str) -> None:
        """Open web page by URL."""
        
        open_url(self._driver, url)

    def url_if_not_current(self, url: str, ignore_trailing_slash: bool = True) -> None:
        """Open a URL if it isn't already the current URL. Useful when doing multiple operations on a page where."""

        open_url_if_not_current(self._driver, url, ignore_trailing_slash)

    def url_in_new_tab(self, url: str) -> None:
        """Open a link in a new browser tab with support across browsers."""

        open_url_in_new_tab(self._driver, url)



from dataclasses import dataclass
from typing import Union
from ..constant import timeout

@dataclass
class LoginCredentials:
    """Object to define a user profile's access credentials.
    
    Separated from the login form so that the multiple users/roles can use the same form."""
    
    username: str
    password: str

@dataclass
class LoginForm:
    """Object to define the login form page.
    
    url: Optional URL to login page.
    
    post_login_url: Optionally await redirect to this URL as a user is typically redirected automatically to a new page or view after logging in.

    post_login_element_xpath: Upon successful login, optionally await this element to be loaded."""
    
    username_input_xpath: str
    password_input_xpath: str
    submit_button_xpath: str
    url: Union[str, None] = None
    post_login_url: Union[str, None] = None
    post_login_element_xpath: Union[str, None] = None
    timeout: int = timeout.DEFAULT



from selenium import webdriver
from .base.driver import BrowserDriver
from .base.type import BrowserType
from ... import helper

class EdgeBrowserDriver(BrowserDriver):
    def ensure_browser_type(self) -> None:
        self.settings.type = BrowserType.EDGE

    def set_webdriver(self) -> object:
        if self.settings.path_to_executable is None:
            return webdriver.Edge(
                options = self.edge_options)
        else:
            return webdriver.Edge(
                executable_path = self.settings.path_to_executable,
                options = self.edge_options)

    def disable_images(self) -> None:
        if self.settings.disable_images:
            self.edge_options.use_chromium = True
            preferences = {"profile.managed_default_content_settings.images": 2, "profile.default_content_settings.images": 2}
            self.edge_options.add_experimental_option("prefs", preferences)

    def enable_headless(self) -> None:
        if self.settings.headless:
            self.edge_options.use_chromium = True
            self.edge_options.add_argument("headless")
            self.edge_options.add_argument("disable-gpu")

    def set_page_load_strategy(self) -> None:
        self.edge_options = helper.set.page_load_strategy(self, self.edge_options)



from abc import ABC, abstractmethod
from selenium.webdriver.chrome.options import Options as ChromeOptions
from selenium.webdriver.edge.options import Options as EdgeOptions
from selenium.webdriver.firefox.options import Options as FirefoxOptions
from selenium.webdriver.firefox.firefox_profile import FirefoxProfile
from selenium.webdriver.ie.options import Options as IEOptions
from selenium.webdriver.safari.options import Options as SafariOptions
from typing import Union
from .settings import BrowserSettings

class BrowserDriver(ABC):
    """Abstract class that contains the Selenium web driver based on browser type and configuration."""
    
    def __init__(self, settings: Union[BrowserSettings, None] = None) -> None:
        """Initiates basic properties of the Selenium web driver."""
        
        self.chrome_options: ChromeOptions = ChromeOptions()
        self.edge_options: EdgeOptions = EdgeOptions()
        self.firefox_options: FirefoxOptions = FirefoxOptions()
        self.firefox_profile: FirefoxProfile = FirefoxProfile()
        self.ie_options: IEOptions = IEOptions()
        self.safari_options: SafariOptions = SafariOptions()

        self.settings = BrowserSettings() if settings is None else settings
        if settings is None:
            self.ensure_browser_type()
        self.set_options_and_profile()
        self.webdriver: object = self.set_webdriver()

    @abstractmethod
    def ensure_browser_type(self) -> None:
        """Method to ensure the correct browser type if a specific browser instance is created directly from a subclass (e.g. FirefoxBrowserDriver) without the optional settings as argument, simply as Chrome is default browser."""
        
        raise NotImplementedError

    @abstractmethod
    def set_webdriver(self) -> object:
        """Method to set web driver based on the settings."""

        raise NotImplementedError

    def set_options_and_profile(self) -> None:
        """Internal task initialiser that runs the configuration methods to disable images, enable headless, etc."""
        
        self.disable_images()
        self.enable_headless()
        self.set_page_load_strategy()

    @abstractmethod
    def disable_images(self) -> None:
        """Method to configure web driver to disable download of images for faster browsing."""
        
        raise NotImplementedError

    @abstractmethod
    def enable_headless(self) -> None:
        """Method to enable headless version of web driver (i.e. don't open browser window) for faster browsing."""

        raise NotImplementedError

    @abstractmethod
    def set_page_load_strategy(self) -> None:
        """Method to set the page load strategy to define whether the web driver should wait until all assets are downloaded (slower) or not (faster)."""

        raise NotImplementedError



from .wait_for_element import wait_for_element
from ..constant import timeout
from ..model.browser.base.driver import BrowserDriver
from ..model.driver_methods import DriverMethods

def tool_count_number_of_elements(driver: object, xpath: str, timeout: int = timeout.DEFAULT) -> int:
	wait_for_element(driver, xpath, timeout)
	elements = driver.find_elements_by_xpath(xpath)
	return len(elements)

class ToolsDriverMethods(DriverMethods):
    def __init__(self, browser_driver: BrowserDriver) -> None:
        super().__init__(browser_driver)

    def count_number_of_elements(self, xpath: str, timeout: int = timeout.DEFAULT) -> int:

        return tool_count_number_of_elements(self._driver, xpath, timeout)



from selenium.common.exceptions import TimeoutException, NoSuchElementException
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait
from ..constant import timeout
from ..exception.element import NoElementFoundException
from ..exception.timeout import WaitForElementTimeoutException

def wait_for_element(driver: object, xpath: str, timeout: int = timeout.DEFAULT) -> None:
    try:
        WebDriverWait(driver, timeout).until(EC.presence_of_element_located((By.XPATH, xpath)))
    except TimeoutException:
        raise WaitForElementTimeoutException(driver, xpath)
    except NoSuchElementException:
        raise NoElementFoundException(driver, xpath)



import time, random
from selenium.common.exceptions import TimeoutException
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait
from typing import List
from .check_if import check_if_is_image_element_loaded
from .check_if_does_element_exist import check_if_does_element_exist
from .wait_for_element import wait_for_element
from .. import helper
from ..constant import timeout
from ..exception.timeout import WaitForPageTitleToChangeTimeoutException, WaitForUrlTimeoutException
from ..model.browser.base.driver import BrowserDriver
from ..model.driver_methods import DriverMethods

def wait_until_element_disappears(driver: object, xpath: str, timeout: int = timeout.DEFAULT) -> None:
    helper.driver.retry_until_condition_is_false(check_if_does_element_exist(driver, xpath), timeout)
    
def wait_until_images_have_loaded(driver: object, xpath: str, timeout: int = timeout.DEFAULT) -> None:
    def are_all_images_loaded(driver: object, elements: List[object]):
        return all(check_if_is_image_element_loaded(driver, element) is not False for element in elements)

    wait_for_element(driver, xpath, timeout)
    elements = driver.find_elements_by_xpath(xpath)
    helper.driver.retry_until_condition_is_true(are_all_images_loaded(driver, elements), timeout)

def wait_until_page_title_contains(driver: object, page_title_fragment: str, timeout: int = timeout.DEFAULT) -> None:
    try:
        WebDriverWait(driver, timeout).until(EC.title_contains(page_title_fragment))
    except TimeoutException:
        raise WaitForPageTitleToChangeTimeoutException(driver, page_title_fragment)

def wait_until_page_title_is(driver: object, page_title: str, timeout: int = timeout.DEFAULT) -> None:
    try:
        WebDriverWait(driver, timeout).until(EC.title_is(page_title))
    except TimeoutException:
        raise WaitForPageTitleToChangeTimeoutException(driver, page_title)

def wait_until_url_contains(driver: object, url: str, timeout: int = timeout.LONG) -> None:
    try:
        WebDriverWait(driver, timeout).until(EC.url_contains(url))
    except TimeoutException:
        raise WaitForUrlTimeoutException(driver, url)

def wait_random_time(min_seconds: int = 1, max_seconds: int = timeout.DEFAULT) -> None:
    time.sleep(random.uniform(min_seconds, max_seconds))

class WaitDriverMethods(DriverMethods):
    def __init__(self, browser_driver: BrowserDriver) -> None:
        super().__init__(browser_driver)

    def for_element(self, xpath: str) -> None:
        """Wait until element is ready in the DOM and/or on the screen.

        Especially useful for single-page app elements handled/modified by JavaScript, but also standard HTML that doesn't load immediately, this helper function ensures that DOM elements are ready before processing."""

        wait_for_element(self._driver, xpath)

    def until_element_disappears(self, xpath: str, timeout: int = timeout.DEFAULT) -> None:
        """Wait until element doesn't exist."""

        wait_until_element_disappears(self._driver, xpath, timeout)

    def until_images_have_loaded(self, xpath: str, timeout: int = timeout.DEFAULT) -> None:
        """Wait until element doesn't exist. The image XPath can target one or more images."""

        wait_until_images_have_loaded(self._driver, xpath, timeout)

    def until_page_title_contains(self, page_title_fragment: str, timeout: int = timeout.DEFAULT) -> None:
        """Wait until the page title has changed, e.g. after a redirect or update. The input can contain both a fragment or the full page title."""

        wait_until_page_title_contains(self._driver, page_title_fragment, timeout)

    def until_page_title_is(self, page_title: str, timeout: int = timeout.DEFAULT) -> None:
        """Wait until the page title has changed, e.g. after a redirect or update. The input has to match the exact page title."""

        wait_until_page_title_contains(self._driver, page_title, timeout)

    def until_url_contains(self, url: str, timeout: int = timeout.LONG) -> None:
        """Wait until the browser URL has changed, e.g. after a redirect. The URL variable can contain both a fragment (e.g. ?login=true) or a full URL (e.g. https://www.example.com/?login=true)."""

        wait_until_url_contains(self._driver, url, timeout)

    def random_time(self, min_seconds: int = 1, max_seconds: int = 5) -> None:
        """Randomize sleep timing to make actions look less like a bot."""

        wait_random_time(min_seconds, max_seconds)



from . import BrowserExtension
from .... import helper

class InternetExplorerBrowserExtension(BrowserExtension):
    def enable_images(self) -> None:
        """Intended use: Revert disabled images for Internet Explorer as its web driver doesn't support a session-based configuration for Selenium, but only a global configuration that impacts normal use."""

        helper.internet_explorer.enable_images(self.browser_driver)



from selenium.common.exceptions import NoSuchElementException

def check_if_does_element_exist(driver: object, xpath: str) -> bool:
    try:
        element = driver.find_element_by_xpath(xpath)
        return element is not None
    except NoSuchElementException:
        return False



from ..model.browser.base.type import BrowserType

class CookieBannerException(Exception):
    def __init__(self) -> None:
        self.message = "Something went wrong when trying to accept cookies. Please try again..."
        super().__init__(self.message)

    def __str__(self) -> str:
        return self.message



import time
from .click import click_button
from .input import input_value
from .open import open_url_if_not_current
from .wait import wait_until_element_disappears, wait_until_url_contains
from .wait_for_element import wait_for_element
from ..constant import timeout
from ..exception.cookie_banner import CookieBannerException
from ..exception.login import LoginException
from ..model.browser.base.driver import BrowserDriver
from ..model.cookie_banner import CookieBannerSettings
from ..model.driver_methods import DriverMethods
from ..model.login import LoginCredentials, LoginForm

def combo_cookie_banner(driver: object, settings: CookieBannerSettings) -> None:
    try:
        if settings.url is not None:
            open_url_if_not_current(driver, settings.url)
        time.sleep(settings.has_loaded_wait_seconds)
        if settings.has_loaded_xpath is not None:
            wait_for_element(driver, settings.has_loaded_xpath)
        click_button(driver, settings.button_xpath)
        time.sleep(settings.has_disappeared_wait_seconds)
        wait_until_element_disappears(driver, settings.button_xpath)
    except Exception:
        raise CookieBannerException()

def combo_login(driver: object, login_credentials: LoginCredentials, login_form: LoginForm, wait_seconds: int = timeout.DEFAULT) -> None:
    try:
        if login_form.url is not None:
            open_url_if_not_current(driver, login_form.url)
        input_value(driver, login_form.username_input_xpath, login_credentials.username)
        input_value(driver, login_form.password_input_xpath, login_credentials.password)
        click_button(driver, login_form.submit_button_xpath)
        time.sleep(wait_seconds)
        if login_form.post_login_url is not None:
            wait_until_url_contains(driver, login_form.post_login_url)
        if login_form.post_login_element_xpath is not None:
            wait_for_element(driver, login_form.post_login_element_xpath)
    except Exception:
        raise LoginException(login_credentials.username)

class ComboDriverMethods(DriverMethods):
    def __init__(self, browser_driver: BrowserDriver) -> None:
        super().__init__(browser_driver)
        
    def cookie_banner(self, settings: CookieBannerSettings) -> None:
        """Standardised combination of methods to accept or decline cookies."""

        combo_cookie_banner(self._driver, settings)

    def login(self, login_credentials: LoginCredentials, login_form: LoginForm, wait_seconds: int = timeout.DEFAULT) -> None:
        """Standardised combination of methods to log in.

        wait_seconds: Extra seconds in addition to timeout to make sure the login is processed and that the user is redirected succesfully."""

        combo_login(self._driver, login_credentials, login_form, wait_seconds)



__all__ = ["Browser", "BrowserSettings", "BrowserType", "CookieBannerSettings", "LoginCredentials", "LoginForm", "PageLoadStrategy", "TimeoutSettings", "TimeoutStrategy"]

from .browser import Browser
from .model.browser.base.page_load_strategy import PageLoadStrategy
from .model.browser.base.settings import BrowserSettings
from .model.browser.base.type import BrowserType
from .model.cookie_banner import CookieBannerSettings
from .model.login import LoginCredentials, LoginForm
from .model.timeout_strategy import TimeoutSettings, TimeoutStrategy
from .version import __version__



import time
from typing import Callable, List, Union
from ..constant import timeout

def retry_and_get_text_from_element(func: Callable[[object, str], str], timeout: int = timeout.DEFAULT, wait_interval_seconds: float = 0.5) -> str:
    text = str(func)
    i = 0
    retries = timeout / wait_interval_seconds
    while not text and i < retries:
        time.sleep(wait_interval_seconds)
        text = str(func)
        i += 1
    return text

def retry_until_condition_is_true(func: Callable[[object, Union[str, List[object]]], bool], timeout: int = timeout.DEFAULT, wait_interval_seconds: float = 0.5) -> None:
    i = 0
    retries = timeout / wait_interval_seconds
    while func is False and i < retries:
        time.sleep(wait_interval_seconds)
        i += 1

def retry_until_condition_is_false(func: Callable[[object, Union[str, List[object]]], bool], timeout: int = timeout.DEFAULT, wait_interval_seconds: float = 0.5) -> None:
    i = 0
    retries = timeout / wait_interval_seconds
    while func is True and i < retries:
        time.sleep(wait_interval_seconds)
        i += 1
